use bytes::{Bytes, BytesMut};
use futures::{sink::Sink, SinkExt};
use rsip::common::method::Method as SipMethod;
use rsip::headers::header::Header as SipHeader;
use rsip::message::{request::Request, response::Response, SipMessage};
use sdp_rs::lines::media::{MediaType, ProtoType};
use sdp_rs::lines::{attribute::Rtpmap, Attribute};
use sdp_rs::{MediaDescription, SessionDescription};
use std::net::SocketAddr;
use std::str;
use std::time::Duration;
use tokio::net::UdpSocket;
use tokio::sync::mpsc;
use tokio::time::sleep;
use tokio_stream::StreamExt;
use tokio_util::{
    codec::{Decoder, Encoder},
    sync::{PollSendError, PollSender},
    udp::UdpFramed,
};
use tracing::{event, instrument, Level};

const SIP_PORT: u16 = 5060;
const BIND_ADDR: &str = "0.0.0.0"; // for now

struct SipCodec {}

impl Decoder for SipCodec {
    type Item = SipMessage;
    type Error = std::io::Error;

    fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>, Self::Error> {
        if src.is_empty() {
            Ok(None)
        } else {
            // We're assuming we get an entire sip message at once here.
            match SipMessage::try_from(src.as_ref()) {
                Ok(msg) => {
                    src.clear();
                    Ok(Some(msg))
                }
                Err(e) => {
                    src.clear(); // Still clear the buf to be ready for the next packet
                    event!(Level::ERROR, error = %e, "Error decoding SIP message.");
                    Ok(None)
                }
            }
        }
    }
}

impl Encoder<SipMessage> for SipCodec {
    type Error = std::io::Error;

    #[instrument(
        level="debug",
        skip(self, item, dst)
    )]
    fn encode(&mut self, item: SipMessage, dst: &mut BytesMut) -> Result<(), Self::Error> {
        let stuff: Bytes = item.into();
        dst.reserve(stuff.len());
        dst.extend_from_slice(&stuff);
        Ok(())
    }
}

type StdErr<T> = Result<T, Box<dyn std::error::Error>>;

struct Server {}

struct CurrentCall {
    remote: SocketAddr,
    call_id: String,
    req_tx: mpsc::Sender<Request>,
}

impl Server {
    #[instrument]
    async fn run_sip() -> StdErr<()> {
        event!(Level::INFO, "Starting...");
        let socket = UdpSocket::bind(format!("{}:{}", BIND_ADDR, SIP_PORT)).await?;
        let mut framed = UdpFramed::new(socket, SipCodec {});

        // State for the current call.
        // TODO: be able to handle multiple SIP sessions at once
        let mut current_call: Option<CurrentCall> = None;
        // Channel for SIP responses from call handlers
        let (response_tx, mut response_rx) = mpsc::channel(10);
        loop {
            tokio::select! {
                // We got a new response to send
                Some((res, send_to)) = response_rx.recv() => {
                    event!(Level::INFO, remote=%send_to, "Sending response");
                    framed.send((SipMessage::from(res), send_to)).await?;
                }
                Some(frame) = framed.next()  => {
                    let (msg, remote) = frame?;
                    match msg {
                        SipMessage::Request(req) => {
                            // TODO: some kind of more general extraction of headers
                            let call_id = req.headers.iter().filter_map(|h| {
                                match h {
                                    SipHeader::CallId(id) => Some(id.clone()),
                                    _ => None,
                                }
                            }).next().ok_or("No CallId header")?.into();
                            event!(Level::INFO, %remote, method=%req.method, %call_id, "Got request");
                            match req.method {
                                SipMethod::Invite => {
                                    // We got a new INVITE. Are we already in a call?
                                    match &current_call {
                                        None => {
                                            // We're not in a call already, guess we are now...
                                            event!( Level::INFO, %remote,  "Starting call");
                                            let (req_tx, req_rx) = mpsc::channel(10);
                                            current_call = Some(CurrentCall { remote, call_id, req_tx });
                                            // Jesus christ. Wrap the response sender to inject the
                                            // remote's address. This means that `handle_call`
                                            // doesn't need to be aware of the remote at all.
                                            let res_tx = response_tx.clone();
                                            tokio::spawn(async move {
                                                let res_tx = PollSender::new(res_tx).with::<_, _, _, PollSendError<_>>(|res| {futures::future::ready(Ok((res, remote)))});
                                                Self::handle_call(&req, res_tx, req_rx).await.unwrap();
                                            });
                                        }
                                        Some(call) => {
                                            // We are in a call.
                                            if call_id != call.call_id {
                                                // If this INVITE is from someone else, say we're
                                                // busy.
                                                event!(
                                                    Level::INFO,
                                                    %remote,
                                                    "Call in progress already, responding with 486 Busy Here",
                                                );
                                                let mut res = Response::default();
                                                res.status_code = 486.into();
                                                response_tx.send((res, remote)).await?;
                                                continue;
                                            } else {
                                                // If it's from the same call... ignore it and warn
                                                // I guess?
                                                event!( Level::WARN, %remote, %call_id, "Got INVITE for same CallID as the one we're currently in");
                                            }

                                        }
                                    }
                                }
                                _ => {
                                    if let Some(call) = &current_call {
                                        if call.call_id == call_id {
                                            // If this request is for the currently handled callId,
                                            // route it to the call handler.
                                            call.req_tx.send(req).await?;
                                        }
                                        event!(Level::WARN, %remote, %call_id, "Ignoring request with different Call ID to the one we're handling");
                                    } else {
                                        event!(Level::WARN, %remote, %call_id, method=%req.method, "Not in a call, ignoring non-INVITE request.");
                                    }
                                }
                            }
                        }
                        SipMessage::Response(_res) => {
                            event!( Level::INFO, %remote, "Got SIP response");
                        }
                    };
                }
            }
        }
    }

    // Check that the media description is something we can handle, returning the selected format
    #[instrument]
    fn select_format(md: &MediaDescription) -> StdErr<(Rtpmap, Option<&str>)> {
        if md.media.media != MediaType::Audio || md.media.proto != ProtoType::RtpAvp {
            return Err("Media is not audio over RTP/AVP".into());
        } else if md.media.num_of_ports != None && md.media.num_of_ports != Some(1) {
            return Err(format!("Media has {:?} ports, want 1", md.media.num_of_ports).into());
        }
        md.attributes
            .iter()
            .filter_map(|a| match a {
                Attribute::Rtpmap(r) => {
                    //TODO: likely need to be smarter here in case of multiple variants of opus
                    if r.encoding_name == "opus" {
                        let prefix = format!("{} ", r.payload_type);
                        // Find the matching fmtp if there is one
                        let fmtp = md.attributes.iter().filter_map(|a| match a{
                            Attribute::Other(fmt, Some(params)) if fmt == "fmtp" => {
                                params.strip_prefix(&prefix)
                            },
                            _ => None
                        }).next();
                        Some((r.clone(), fmtp))
                    } else {
                        None
                    }
                }
                _ => None,
            })
            .next()
            .ok_or("No accepted codecs found".into())
    }

    // Handle a call
    #[instrument(
        level = "info",
        skip(invite, responses, requests)
        fields()
    )]
    async fn handle_call<T: Sink<Response> + std::marker::Unpin>(
        invite: &Request,
        mut responses: T,
        mut requests: mpsc::Receiver<Request>,
    ) -> StdErr<()> {
        let mut base_res = Response::default();
        // Copy headers from the invite
        base_res.headers.extend(
            invite
                .headers
                .iter()
                .filter(|h| {
                    matches!(
                        h,
                        SipHeader::CSeq(_)
                            | SipHeader::CallId(_)
                            | SipHeader::From(_)
                            | SipHeader::To(_)
                            | SipHeader::Via(_)
                    )
                })
                .cloned()
                .collect(),
        );
        // 100 trying...
        let mut res = base_res.clone();
        res.status_code = 100.into();
        responses.send(res).await;

        let body = str::from_utf8(invite.body().as_ref())?;
        let sdp = SessionDescription::try_from(body)?;
        let (md, rtpmap, fmtp) = sdp
            .media_descriptions
            .iter()
            .filter_map(|md| match Self::select_format(md) {
                Ok((rtpmap, fmtp)) => Some((md, rtpmap, fmtp)), 
                Err(e) => {
                    event!(Level::INFO, why=e, media_description=%md, "Rejected media description.");
                    None
                }
            })
            .next()
            .ok_or("No supported media descriptions")?;
        event!(Level::INFO, media_desc=%md, %rtpmap, ?fmtp, "Selected rtpmap/codec");

        // Setup RTP. Bind to an os-allocated port
        let socket = UdpSocket::bind(format!("{}:0", BIND_ADDR)).await?;
        let rtp_port = socket.local_addr()?.port();
        event!(Level::INFO, rtp_port, "Bound RTP port");
        let mut framed = UdpFramed::new(socket, SipCodec {});

        loop {
            tokio::select! {
                Some(req) = requests.recv() => {
                    event!(Level::INFO, ?req, "Got request");
                    match req.method {
                        SipMethod::Bye => {
                            event!(Level::INFO, "Remote has hung up.");
                            //TODO: Cancel RTP streams and clean up
                            break
                        },
                        _ => {event!(Level::WARN, method=%req.method, "Call handler got unimplemented SIP method!");}
                    }

                },
                _ = sleep(Duration::from_millis(3000)) => {
                    // Reject call after 3000 seconds
                    let mut res = base_res.clone();
                    res.status_code = 603.into();
                    responses.send(res).await;
                    break
                }
            }
        }

        event!(Level::INFO, "Call handler loop done.");

        Ok(())
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    tracing_subscriber::fmt::init();
    Server::run_sip().await
}
